/*
 * Copyright (c) 2004, University of Illinois,Distributed Systems Group
 * Large & Dynamic Groups Size Estimation project members:
 * Dimitrios Psaltoulis
 * Dionysios Kostoulas
 * Indranil Gupta
 * Al Demers
 * Ken Birman
 *  All rights reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * - Neither the name of the University of Illinois, Urbana-Champaign
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * HopCountingAlgorithm.java
 *
 * Created on June 17, 2004, 12:55 PM
 */

package SizeEstimationAlgorithm.HopCounting;

import java.net.*;
import java.util.*;

import SizeEstimationAlgorithm.*;
/**
 *
 * @author  dimitrios psaltoulis
 */
public class HopCountingAlgorithm extends GossipAlgorithm {
    
    public HopCountingAlgorithm(InetSocketAddress[] members) {
        this(members, GossipAlgorithm.DEFAULT_PORT);
    }
    
    public HopCountingAlgorithm(InetSocketAddress[] members, int serverPort) {
        this.members = members;
        this.serverPort = serverPort;
        server = new Server(serverPort, this);
        server.start();
        if (GossipAlgorithm.DEBUG) System.out.println("Hop Counting Algorithm initialized\n");
    }
    
    public String getKey(GossipId g) {
        return new String(g.getIP().toString()+g.getPort()+g.getSeqNumber());
    }
    
    public void startCounting() {
        gossipSeqNumber++;
        /*if (GossipAlgorithm.DEBUG)*/ System.out.println("\nStarting gossip with sequence number: " + gossipSeqNumber + "\n");
        try {
            GossipId gossipId = new GossipId(InetAddress.getLocalHost().getHostAddress(), serverPort, gossipSeqNumber);
            GossipMessage gossipMessage =  new GossipMessage(gossipId, 0, false);
            HopGossipInfo gInfo =  new HopGossipInfo(0, 0, 0);
            int i=0,k=0;
            while( (i<GossipAlgorithm.SEND_TO)&&(k<5) ) {
                if(members[k] != null) {
                    sendMessage(members[k], gossipMessage);
                    gInfo.incNumberOfMessagesSent();
                    ++i;
                }
                ++k;
            }
            hashMap.put(getKey(gossipId), gInfo);
        } catch(UnknownHostException e) {
            System.err.println(e);
        }
    }
    
    public void receiveMessage(GossipMessage g) throws UnknownHostException {
        GossipId gId = g.getGossipId();
        String key = getKey(gId);
        if( hashMap.containsKey(key) ) {
            if ( (gId.getIP().equals(InetAddress.getLocalHost())) && (gId.getPort()==serverPort) ) {
                HopGossipInfo gInfo = (HopGossipInfo)hashMap.get(key);
                if(g.getReport()) {
                    gInfo.addReport(g.getHopDistance());
                    size = gInfo.estimateSizeAv(GossipAlgorithm.SEND_TO, MIN_HOP_REPORT);
                    sizeS = gInfo.estimateSizeSample(GossipAlgorithm.SEND_TO, MIN_HOP_REPORT);
                    if (GossipAlgorithm.DEBUG) System.out.println("Report msg.");
                    /*if (GossipAlgorithm.DEBUG)*/ System.out.println("Gossip sequence number: " + gId.getSeqNumber());
                    if (GossipAlgorithm.DEBUG) System.out.println("Hop distance: " + g.getHopDistance());
                    /*if (GossipAlgorithm.DEBUG)*/ System.out.println("Network size: " + size);
                    /*if (GossipAlgorithm.DEBUG)*/ System.out.println("Network size(S): " + sizeS +"\n");
                }
                else {
                    gInfo.incNumberOfMessagesReceived();
                    if (GossipAlgorithm.DEBUG) System.out.println("Old gossip(mine).\n");
                    hashMap.put(key, gInfo);
                }
            }
            else {
                g.incHopDistance();
                HopGossipInfo gInfo = (HopGossipInfo)hashMap.get(key);
                gInfo.incNumberOfMessagesReceived();
                gInfo.setHopDistance( Math.min( gInfo.getHopDistance(), g.getHopDistance()) );
                if (GossipAlgorithm.DEBUG) System.out.println("Old gossip.\nDistance from init: " + gInfo.getHopDistance() + "\n");
                hashMap.put(key, gInfo);
            }
        }
        else {
            g.incHopDistance();
            int hDist = g.getHopDistance();
            HopGossipInfo gInfo = new HopGossipInfo(hDist, 1, 0);
            int i=0,k=0;
            while( (i<GossipAlgorithm.SEND_TO)&&(k<5) ) {
                if(members[k]!= null) {
                    sendMessage(members[k], g);
                    gInfo.incNumberOfMessagesSent();
                    ++i;
                }
                ++k;
            }
            hashMap.put(key, gInfo);
            if (GossipAlgorithm.DEBUG) System.out.println("New gossip.\nDistance from init: " + hDist + "\n");

            if( (hDist<=MIN_HOP_REPORT)
            || ( (hDist>MIN_HOP_REPORT)&&(Math.random()<Math.pow(GossipAlgorithm.SEND_TO, (MIN_HOP_REPORT-hDist))) ) ){
                g.setReport(true);
                sendMessage(g.getGossipId().getAddress(), g);
            }
        }
    }
    
    public long getSize() { return size; }
    public long getSizeS() { return sizeS; }
    
    public final int MIN_HOP_REPORT = 4;
    
    private int serverPort;
    
    private long size=0;
    private long sizeS=0;
    private InetSocketAddress[] members = new InetSocketAddress[GossipAlgorithm.MAX_PEERS];
    private Map hashMap = new HashMap(100, (float)0.75);
    private int gossipSeqNumber;
}