/*
 * Copyright (c) 2004, University of Illinois,Distributed Systems Group
 * Large & Dynamic Groups Size Estimation project members:
 * Dimitrios Psaltoulis
 * Dionysios Kostoulas
 * Indranil Gupta
 * Al Demers
 * Ken Birman
 *  All rights reserved.
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the
 * distribution.
 *
 * - Neither the name of the University of Illinois, Urbana-Champaign
 * nor the names of its contributors may be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * IntervalCountingAlgorithm.java
 *
 * Created on June 17, 2004, 1:05 PM
 */

package SizeEstimationAlgorithm.IntervalCounting;


import java.net.*;
import java.security.*;
import java.util.*;

import SizeEstimationAlgorithm.*;
import SizeEstimationAlgorithm.IntervalCounting.utils.*;
/**
 *
 * @author  dionysios kostoulas
 */
// MAXMEMORY threshold is used for changing the size of the interval
// A maximum of MAXMEMORY peers can be remembered. After the threshold
// is met the size of the interval stops increasing.
// Estimates Converge is to be implemented as an alternative method
// for changing the size of the interval in a future version.
// A single selection method is used for setting the centre of the
// interval. Periodically changing selection methods and multiple
// runs methods, which have been found to improve the performance of algorithm,
// are to be implemented in a future version.

public class IntervalCountingAlgorithm extends GossipAlgorithm {
    
    public IntervalCountingAlgorithm(InetSocketAddress[] members) {
        this(members, GossipAlgorithm.DEFAULT_PORT);
    }
    
    public IntervalCountingAlgorithm(InetSocketAddress[] members, int serverPort) {
        this.members = members;
        this.serverPort = serverPort;
        server = new Server(serverPort, this);
        server.start();
        try {
            pId = new HashId(shaHash(InetAddress.getLocalHost(), serverPort));
        } catch (UnknownHostException e) { }
        peerList = new MemberList();
        interval = new Interval(20,19,21);
        round = 0;
        if (GossipAlgorithm.DEBUG) System.out.println("Interval Density Counting Algorithm initialized\n");
    }
    
    // Convert the String argument (IP address + port number)
    // to a SHA-1 hash value, represented as a hex string
    
    public String shaHash(InetAddress ip, int port) {
        String s = new StringBuffer().append(ip).append(port).toString();
        
        MessageDigest md = null;
        
        try {
            md = MessageDigest.getInstance("SHA");
        }
        catch(NoSuchAlgorithmException e) {
            System.out.println("No such algorithm");
        }
        
        byte[] byteArray = s.getBytes();
        
        byte[] hashArray = md.digest(byteArray);
        
        //Convert hash array to hex
        String hexHash = "";
        
        for(int i = 0; i < hashArray.length; i++) {
            int val=hashArray[i];
            if(val<0) val+=256;
            int sym1 = val%16;
            int sym2 = val/16;
            if(sym2<10) hexHash+= (char)((int)'0'+ sym2);
            else hexHash+=(char)((int)'A'+(sym2 - 10));
            if(sym1<10) hexHash+=(char)((int)'0'+ sym1);
            else hexHash+=(char)((int)'A'+sym1 - 10);
        }
        
        return hexHash.substring(0,2*20);
    }
    
    public void startCounting() {
        gossipSeqNumber++;
        receives = 0;
        
        /*if (GossipAlgorithm.DEBUG)*/ System.out.println("\nStarting gossip with sequence number: " + gossipSeqNumber + "\n");
        try {
            GossipId gossipId = new GossipId(InetAddress.getLocalHost(), serverPort, gossipSeqNumber);
            GossipMessage gossipMessage =  new GossipMessage(gossipId, peerList, interval);
            if (pId.isWithin(interval) && (peerList.contains(pId)==-1)) {
                peerList.add(new ListEntry(pId, heartbeat, round));
            }
            report = new ReportEntry(gossipId, peerList.getSize(), interval.getSize());
            int i=0,k=0;
            while( (i<GossipAlgorithm.SEND_TO)&&(k<5) ) {
                if(members[k]!=null) {
                    sendMessage(members[k], (Object)gossipMessage);
                    ++i;
                }
                ++k;
            }
        } catch(UnknownHostException e) {
            System.err.println(e);
        }
    }
    
    public void receiveMessage(GossipMessage g) throws UnknownHostException {
        GossipId gId = g.getGossipId();
        peerList.mergeWith(g.getPeerList());
        //peerList.clear(round);
        if (!( (gId.getIP().equals(InetAddress.getLocalHost()))
        && (gId.getPort()==serverPort) )) {
            interval = g.getInterval();
        }
        if (pId.isWithin(interval) && peerList.contains(pId)==-1) {
            peerList.add(new ListEntry(pId, heartbeat+1, round));
        }
        g.setPeerList(peerList);
        if ( (gId.getIP().equals(InetAddress.getLocalHost()))
        && (gId.getPort()==serverPort) ) {
            if (maxSeqNumber <= gId.getSeqNumber()) {
                maxSeqNumber = gId.getSeqNumber();
            }
            round++;
            sizeEstimate = peerList.getSize() * (40 / interval.getSize());
            if ((peerList.getSize() < peerList.maxmemory) && (interval.getSize() < 40) && (gId.getSeqNumber() == gossipSeqNumber) ) {
                interval.changeSize(2);
                if (GossipAlgorithm.DEBUG) System.out.println("\n" + "Increasing interval..." + "\n");
                startCounting();
            }
            else {
                if (maxSeqNumber <= gId.getSeqNumber()) {
                    reportResults(0, gId);
                }
                if (gossipSeqNumber == maxSeqNumber) {
                    if (sizeEstimate >= maxEstimate) {
                        reportResults(1, gId);
                        maxEstimate = sizeEstimate;
                    }
                }
            }
        }
        else {
            if (maxSeqNumber < gId.getSeqNumber()) {
                report = new ReportEntry(gId, g.getPeerList().getSize(), g.getInterval().getSize());
                maxSeqNumber = gId.getSeqNumber();
                receives = 0;
            }
            
            if (((receives > 0)
            && (report.gossipId.getIP().equals(gId.getIP()))
            && (report.gossipId.getPort() == gId.getPort())
            && (report.gossipId.getSeqNumber() == gId.getSeqNumber())
            && (report.listSize == g.getPeerList().getSize()) && (report.intervalSize == g.getInterval().getSize()))
            || ((receives > 0) && (maxSeqNumber > gId.getSeqNumber()))) {
                if (GossipAlgorithm.DEBUG) System.out.println("Old gossip received" + "\n");
            }
            else {
                receives++;
                if (receives > 1) report.update(gId, g.getPeerList().getSize(), g.getInterval().getSize());
                if (GossipAlgorithm.DEBUG) System.out.println("New gossip received" + "\n");
                
                int i=0,k=0;
                while( (i<GossipAlgorithm.SEND_TO)&&(k<5) ) {
                    if(members[k]!=null){
                        sendMessage(members[k], (Object)g);
                        ++i;
                    }
                    ++k;
                }
            }
        }
        
    }
    
    public void reportResults(int mode, GossipId gId) {
        if (mode == 0) {
            if (GossipAlgorithm.DEBUG) System.out.println("Gossip sequence number: " + gId.getSeqNumber()+ "\n" +
            "Network size estimate: " + sizeEstimate + "\n");
        }
        else if (mode == 1) {
            System.out.println("Network size estimation report" + "\n" + "Gossip sequence number: " + gId.getSeqNumber()+ "\n" +
            "Network size estimate: " + sizeEstimate + "\n");
        }
    }

    public long getSize() { return sizeEstimate; }
    
    private int serverPort;
    private static InetSocketAddress[] members  = new InetSocketAddress[GossipAlgorithm.MAX_PEERS];
    private HashId pId;
    private MemberList peerList;
    private Interval interval;
    private ReportEntry report;
    private int sizeEstimate = 0;
    private int maxEstimate = 0;
    private int heartbeat = 0;
    private int round;
    private int receives;
    private int gossipSeqNumber;
    private int maxSeqNumber = 0;
}